#include <string>
#include <iostream>
#include "cobra-ast.hh"
#include "cobra-assem.hh"

//
// Prgm::compile(os):
//
// Compile a Cobra program, namely its series of function definitions
// (including its "main" function) into MIPS assembly code, outputting
// it to the `os` stream.
//
// The compiler works in three passes:
//
// PASS 1: Assemble the code of each function definition into
//         sequences of "pseudo-instructions". Each pseudo instruction
//         is an object of subtype of INST. Each operates on
//         temporaries corresponding to variables used in the Cobra
//         code. It also introduces new temporaries to break down
//         Cobra expressions and statements into several
//         pseudo-instructions.
//
// PASS 2: For each assembled function component, assign stack frame
//         slots to each of the temporaries used within its sequence
//         of pseudo-instructions.
//
// PASS 3: output MIPS code that is equivalent to the assembled
//         pseudo-instructions. For each function this includes ENTER
//         and LEAVE code that sets up and takes down the stack
//         frame. Within that code, temporaries are loaded from, and
//         stored to, their allocated slot within the stack frame.
//
void Prgm::compile(std::ostream& os) const {
  Assembly assm { };
  assm.dataString("eoln","\\n");
  assemble(assm);
  assm.frameVars();
  assm.toMIPS(os);
}

// Prgm::assemble(assm):
//
// This is a helper method used by compile. It assembles
// pseudo-instruction sequences for each function definition.
//
void Prgm::assemble(Assembly& assm) const {
  for (Defn_ptr defn : defns) {
    assm.addComponent(defn);
    defn->assemble(assm);
  }
}

// Defn::assemble(assm):
//
// This assembles a sequence of pseudo-instructions that execute the
// body of a function definition.
//
void Defn::assemble(Assembly& assm) const {
  assm.append(INST_ptr {new ENTER {}});
  body->assemble(assm);
  assm.append(INST_ptr {new LEAVE {}});
}

void Updt::assemble(Assembly& assm) const {
  // Your code goes here! Exercise 6.2
}

void Asgn::assemble(Assembly& assm) const {
  valu->assembleToSet(name,assm);
}

void Pass::assemble(Assembly& assm) const {
  assm.append(INST_ptr {new NOP {}});
}

void Rtrn::assemble(Assembly& assm) const {
  Name tmp = assm.generateTmp();
  valu->assembleToSet(tmp,assm);
  assm.append(INST_ptr {new RTN {tmp}});
}

void ForL::assemble(Assembly& assm) const {
  // Your code goes here! Exercise 6.6
}

void Whle::assemble(Assembly& assm) const {
  Label loop_lbl = assm.generateLabel("loop");
  Label endloop_lbl = assm.generateLabel("endloop");
  assm.append(INST_ptr {new LBL {loop_lbl}});  
  test->assembleToBranchIfFalse(endloop_lbl,assm);
  body->assemble(assm);
  assm.append(INST_ptr {new BRA {loop_lbl}});
  assm.append(INST_ptr {new LBL {endloop_lbl}});
}

void IfTE::assemble(Assembly& assm) const {
  Label if_lbl = assm.generateLabel("if");
  Label else_lbl = assm.generateLabel("else");
  Label endif_lbl = assm.generateLabel("endif");
  assm.append(INST_ptr {new LBL {if_lbl}});  
  test->assembleToBranchIfFalse(else_lbl,assm);
  tbdy->assemble(assm);
  assm.append(INST_ptr {new BRA {endif_lbl}}); 
  assm.append(INST_ptr {new LBL {else_lbl}});
  ebdy->assemble(assm);
  assm.append(INST_ptr {new LBL {endif_lbl}});
}

void IfTn::assemble(Assembly& assm) const {
  // Your code goes here! Exercise 6.1
}

void PrtI::assemble(Assembly& assm) const {
  Name val = assm.generateTmp();
  valu->assembleToSet(val,assm);
  assm.append(INST_ptr {new PUTI {val}});
}

void PrtS::assemble(Assembly& assm) const {
  Label str_lbl = assm.generateLabel("print");
  assm.dataString(str_lbl,outs);
  assm.append(INST_ptr {new PUTS {str_lbl}});
}

void Blck::assemble(Assembly& assm) const {
  for (Stmt_ptr stmt : stmts) {
    stmt->assemble(assm);
  }
}

void PCll::assemble(Assembly& assm) const {
  int argi = 0;
  for (Expn_ptr argm : argms) {
    Name tmp = assm.generateTmp("arg"+std::to_string(argi));
    argm->assembleToSet(tmp,assm);
    assm.append(INST_ptr {new ARG {argi,tmp}});
    argi++;
  }
  assm.append(INST_ptr {new CLL {name}});
}

//
// Expn::assembleToSet(dst,assm):
//
// This method assembles an expression tree into a series of
// pseudo-instuctions that compute the value of that expression.
// The value it computes gets placed into the temporary named
// `dst`.
//
// Since expression trees generally have several sub-expressions,
// and since pseudo-instructions operate on only a few values,
// the assembly of an expression generates temporaries that store
// the computed results of its subexpressions. It does this using
// calls to `generateTmp`. This introduces new, machine-generated
// (and unique), names throughout the assembled code.
//
// There is an `assembleToSet` method defined for each expression
// type, that is, every class derived from class Expn.
//
//
void Plus::assembleToSet(Name dst, Assembly& assm) const {

  // Invent two new temporary names.
  Name src1 = assm.generateTmp();
  Name src2 = assm.generateTmp();
  
  // Assemble the left subexpression to set the first.
  left->assembleToSet(src1, assm);
  // Assemble the right subexpression to set the second.
  rght->assembleToSet(src2, assm);

  // Append an ADD instruction that sums them into `dst`.
  assm.append(INST_ptr {new ADD {dst,src1,src2}});
}
//
void Mnus::assembleToSet(Name dst, Assembly& assm) const {
  // Your code goes here! Exercise 6.3
}
//
void Tmes::assembleToSet(Name dst, Assembly& assm) const {
  // Invent two new temporary names.
  Name src1 = assm.generateTmp();
  Name src2 = assm.generateTmp();
  
  // Assemble the left subexpression to set the first.
  left->assembleToSet(src1, assm);
  // Assemble the right subexpression to set the second.
  rght->assembleToSet(src2, assm);

  // Append a MUL instruction, putting their product into `dst`.
  assm.append(INST_ptr {new MUL {dst,src1,src2}});
}
//
void Nmbr::assembleToSet(Name dst, Assembly& assm) const {
  // Set the `dst` to the integer value.
  assm.append(INST_ptr {new SET {dst,valu}});
}
//
void Lkup::assembleToSet(Name dst, Assembly& assm) const {
  // Move the temporary into `dst`.
  assm.append(INST_ptr {new MOV {dst,name}});
}
//
void Inpt::assembleToSet(Name dst, Assembly& assm) const {

  //
  // Create a labelled string constant for the input prompt
  // into the assembled data segment.
  //
  Label str_lbl = assm.generateLabel("input");
  assm.dataString(str_lbl,prpt);

  // Output the prompt string.
  assm.append(INST_ptr {new PUTS{str_lbl}});

  // Get an integer input.
  assm.append(INST_ptr {new GET{dst}});
}
//
void Rndm::assembleToSet(Name dst, Assembly& assm) const {

  // Request a random number.
  assm.append(INST_ptr {new RND{dst}});
}
//
void FCll::assembleToSet(Name dst, Assembly& assm) const {
  //
  // Assemble each expression being passed as an argument to the
  // function being called, generating a temporary for each.
  //
  int argi = 0;
  for (Expn_ptr argm : argms) {
    Name tmp = assm.generateTmp();
    argm->assembleToSet(tmp,assm);
    // Set the i-th argument with the `ARG` pseudo-instruction.
    assm.append(INST_ptr {new ARG {argi,tmp}});
    argi++;
  }
  
  // Issue a call pseudo-instruction.
  assm.append(INST_ptr {new CLL {name}});
  
  // Get the returned value, storing it in `dst`.
  assm.append(INST_ptr {new RTV {dst}});
}

//
// Cndn::assembleToBranchIfTrue(tlbl,assm):
//
// This method assembles the code that checks the condition and jumps
// to the given label if the result is true. It falls through to the
// next line of the result is false.
//
// Cndn::assembleToBranchIfFalse(flbl,assm):
//
// This method assembles the code that checks the condition and jumps
// to the given label if the result is false. It falls through to the
// next line of the result is true.
//
//
void Less::assembleToBranchIfTrue(Label tlbl,Assembly& assm) const {
  Name src1 = assm.generateTmp();
  Name src2 = assm.generateTmp();
  left->assembleToSet(src1,assm);
  rght->assembleToSet(src2,assm);
  assm.append(INST_ptr {new BLT {src1,src2,tlbl}});
}
//		
void Less::assembleToBranchIfFalse(Label flbl,Assembly& assm) const {
  Name src1 = assm.generateTmp();
  Name src2 = assm.generateTmp();
  left->assembleToSet(src1,assm);
  rght->assembleToSet(src2,assm);
  assm.append(INST_ptr {new BGE {src1,src2,flbl}});
}
//		
void Eqls::assembleToBranchIfTrue(Label tlbl,Assembly& assm) const {
  Name src1 = assm.generateTmp();
  Name src2 = assm.generateTmp();
  left->assembleToSet(src1,assm);
  rght->assembleToSet(src2,assm);
  assm.append(INST_ptr {new BEQ {src1,src2,tlbl}});
}
//
void Eqls::assembleToBranchIfFalse(Label flbl,Assembly& assm) const {
  Name src1 = assm.generateTmp();
  Name src2 = assm.generateTmp();
  left->assembleToSet(src1,assm);
  rght->assembleToSet(src2,assm);
  assm.append(INST_ptr {new BNE {src1,src2,flbl}});
}
//
void LEql::assembleToBranchIfTrue(Label tlbl,Assembly& assm) const {
  // Your code goes here! Exercise 6.4
}
//		
void LEql::assembleToBranchIfFalse(Label flbl,Assembly& assm) const {
  // Your code goes here! Exercise 6.4
  }
//		
void AndC::assembleToBranchIfTrue(Label tlbl,Assembly& assm) const {
  Label lblA = assm.generateLabel("and");
  left->assembleToBranchIfFalse(lblA,assm);
  rght->assembleToBranchIfTrue(tlbl,assm);
  assm.append(INST_ptr {new LBL {lblA}});
}
//
void AndC::assembleToBranchIfFalse(Label flbl,Assembly& assm) const {
  left->assembleToBranchIfFalse(flbl,assm);
  rght->assembleToBranchIfFalse(flbl,assm);
}
//
void OrCn::assembleToBranchIfTrue(Label tlbl,Assembly& assm) const {
  // Your code goes here! Exercise 6.5
}
//
void OrCn::assembleToBranchIfFalse(Label flbl,Assembly& assm) const {
  // Your code goes here! Exercise 6.5
}
//
void NotC::assembleToBranchIfTrue(Label tlbl,Assembly& assm) const {
  cndn->assembleToBranchIfFalse(tlbl,assm);
}
//
void NotC::assembleToBranchIfFalse(Label flbl,Assembly& assm) const {
  cndn->assembleToBranchIfTrue(flbl,assm);
}
//
//


//
// INST::toMIPS(os,assm)
//
// Method for generating MIPS code that performs the work of
// an Assembly pseudo-instruction (an object derived from class
// INST).
//
// The method outputs a series of MIPS instructions to the output
// stream `os`, using information from the assembled program `assem`.
// In particular, it loads variables from the stack frame at their
// assigned offset within the frame to set some MIPS registers and/or
// stores register values to the stack frameif they are being updated.
// 
// We define this method for each subclass of INST.
//
void SET::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tli $t0," << val << std::endl;
  os << "\tsw $t0," << assm.offsetFor(dst) << "($fp)" << std::endl;
}
//
void MOV::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tlw $t1," << assm.offsetFor(src) << "($fp)" << std::endl;
  os << "\tmove $t0,$t1" << std::endl;
  os << "\tsw $t0," << assm.offsetFor(dst) << "($fp)" << std::endl;
}
//
void RTV::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tmove $t0,$v0" << std::endl;
  os << "\tsw $t0," << assm.offsetFor(dst) << "($fp)" << std::endl;
}
//
void GET::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tli $v0,5" << std::endl;
  os << "\tsyscall" << std::endl;
  os << "\tsw $v0," << assm.offsetFor(dst) << "($fp)" << std::endl;
}
//
void RND::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tli $t0,42" << std::endl; // We only generate 42.
  os << "\tsw $t0," << assm.offsetFor(dst) << "($fp)" << std::endl;
}
//
void NOP::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tmove $v0,$v0" << std::endl;
}
//
void PUTI::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tlw $a0," << assm.offsetFor(src) << "($fp)" << std::endl;
  os << "\tli $v0,1" << std::endl;
  os << "\tsyscall" << std::endl;
  os << "\tli $v0,4" << std::endl;
  os << "\tla $a0,eoln" << std::endl;
  os << "\tsyscall" << std::endl;
}
//
void PUTS::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tli $v0,4" << std::endl;
  os << "\tla $a0," << lbl << std::endl;
  os << "\tsyscall" << std::endl;
  os << "\tli $v0,4" << std::endl;
  os << "\tla $a0,eoln" << std::endl;
  os << "\tsyscall" << std::endl;
}
//
void ADD::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tlw $t1," << assm.offsetFor(src1) << "($fp)" << std::endl;
  os << "\tlw $t2," << assm.offsetFor(src2) << "($fp)" << std::endl;
  os << "\tadd $t0,$t1,$t2" << std::endl;
  os << "\tsw $t0," << assm.offsetFor(dst) << "($fp)" << std::endl;
}
//
void INC::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tlw $t1," << assm.offsetFor(dst) << "($fp)" << std::endl;
  os << "\tlw $t2," << assm.offsetFor(src) << "($fp)" << std::endl;
  os << "\tadd $t0,$t1,$t2" << std::endl;
  os << "\tsw $t0," << assm.offsetFor(dst) << "($fp)" << std::endl;
}
//
void SUB::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tlw $t1," << assm.offsetFor(src1) << "($fp)" << std::endl;
  os << "\tlw $t2," << assm.offsetFor(src2) << "($fp)" << std::endl;
  os << "\tsub $t0,$t1,$t2" << std::endl;
  os << "\tsw $t0," << assm.offsetFor(dst) << "($fp)" << std::endl;
}
//
void MUL::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tlw $t1," << assm.offsetFor(src1) << "($fp)" << std::endl;
  os << "\tlw $t2," << assm.offsetFor(src2) << "($fp)" << std::endl;
  os << "\tmul $t0,$t1,$t2" << std::endl;
  os << "\tsw $t0," << assm.offsetFor(dst) << "($fp)" << std::endl;
}
//
void RTN::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tlw $v0," << assm.offsetFor(src) << "($fp)" << std::endl;
  os << "\tb " << assm.currentFunc()->name << "_return" << std::endl;
}
//
void BCN::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tlw $t1," << assm.offsetFor(src1) << "($fp)" << std::endl;
  os << "\tlw $t2," << assm.offsetFor(src2) << "($fp)" << std::endl;
  os << "\tb" << cndn << " $t1,$t2," << lbl << std::endl;
}
//
void BRA::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tb " << lbl << std::endl;
}
//
void CLL::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tjal " << lbl << std::endl;
}
//
void LBL::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << lbl << ":" << std::endl;
}
//
void ARG::toMIPS(std::ostream& os, const Assembly& assm) const {
  os << "\tlw $a" << idx << "," << assm.offsetFor(src) << "($fp)" << std::endl;
}
//
void ENTER::toMIPS(std::ostream& os, const Assembly& assm) const {
  int ra_slot = assm.offsetFor("saved_return_address");
  int fp_slot = assm.offsetFor("saved_frame_pointer");
  os << assm.currentFunc()->name << ":" << std::endl;
  os << "\tsw $ra," << ra_slot << "($sp)" << std::endl;
  os << "\tsw $fp," << fp_slot << "($sp)" << std::endl;
  os << "\tmove $fp, $sp" << std::endl;
  os << "\taddi $sp,$sp,-" << assm.frameSize() << std::endl;
  int argi = 0;
  for (Name pram : assm.currentFunc()->prams) {
    int slot = assm.offsetFor(pram);
    os << "\tsw $a" << argi << "," << slot << "($fp)" << std::endl;
    argi++;
  }
}
//
void LEAVE::toMIPS(std::ostream& os, const Assembly& assm) const {
  int ra_slot = assm.offsetFor("saved_return_address");
  int fp_slot = assm.offsetFor("saved_frame_pointer");
  os << assm.currentFunc()->name << "_return:" << std::endl;
  os << "\tlw $ra," << ra_slot << "($fp)" << std::endl;
  os << "\tlw $fp," << fp_slot << "($fp)" << std::endl;
  os << "\taddi $sp,$sp," << assm.frameSize() << std::endl;
  os << "\tjr $ra" << std::endl;
}
//
//


//
// INST::frameVars(assm):
//
// Examine the variables used by a pseudo-instruction. Place
// each in some slot in the frame for the currently assembled
// function. (If the variable has already been placed the call
// to `frameVar` does nothing.
//
// This collectively is used to allocate slots in the frame
// of the assembled code for a function, giving each variable
// a location (an offset) within the frame.
//
// We define the method for each subclass of INST.
//
void SET::frameVars(Assembly& assm) {
  assm.frameVar(dst);
}
//
void MOV::frameVars(Assembly& assm) {
  assm.frameVar(dst);
  assm.frameVar(src);
}
//
void RTV::frameVars(Assembly& assm) {
  assm.frameVar(dst);
}
//
void GET::frameVars(Assembly& assm) {
  assm.frameVar(dst);
}
//
void RND::frameVars(Assembly& assm) {
  assm.frameVar(dst);
}
//
void PUTI::frameVars(Assembly& assm) {
  assm.frameVar(src);
}
//
void ADD::frameVars(Assembly& assm) {
  assm.frameVar(dst);
  assm.frameVar(src1);
  assm.frameVar(src2);
}
//
void INC::frameVars(Assembly& assm) {
  assm.frameVar(dst);
  assm.frameVar(src);
}
//
void SUB::frameVars(Assembly& assm) {
  assm.frameVar(dst);
  assm.frameVar(src1);
  assm.frameVar(src2);
}
//
void MUL::frameVars(Assembly& assm) {
  assm.frameVar(dst);
  assm.frameVar(src1);
  assm.frameVar(src2);
}
//
void RTN::frameVars(Assembly& assm) {
  assm.frameVar(src);
}
//
void BCN::frameVars(Assembly& assm) {
  assm.frameVar(src1);
  assm.frameVar(src2);
}
//
void ARG::frameVars(Assembly& assm) {
  assm.frameVar(src);
}
//
//
//

