#ifndef _COBRA_ASSEM_H
#define _COBRA_ASSEM_H

//
// Objects used for compilation and assembly of a Cobra program
// to a MIPS program.
//
// It defines two kinds of object class:
//
//  * INST subclasses - defines pseudo-instructions that form the
//                      intermediate representation of a Cobra
//                      program during its staged conversion to a
//                      MIPS program.
//
//  * Assembly and AssemblyComponent - book-keeping information
//                      and assembled pseudo-instructions for
//                      the compiled Cobra program's components.
//
//

#include <vector>
#include <utility>
#include <string>
#include <memory>
#include "cobra-ast.hh"

class Assembly; // Needed because of the circularity of this
                // class's with each INST subclass's definition.

//
//
// ************************************************************

//
// class INST
//
// This is the abstract class for all pseudo-instructions.
// 
// A sequence of pseudo-instructions forms the intermediate
// representation for the staged conversion of a Cobra program's code
// to its assembled MIPS code.
//
// Pseudo-instructions are, for the most part, very nearly the same as
// MIPS machine instructions except, rather than operating on MIPS
// registers, the can operate on arbirary "temporary" variables. And
// so their work isn't directly tied to a specific processor.
//
// Below this class definition are derived subclasses for the variety
// of pseudo-instructions, like ADD, SET, BRA, etc. that directly
// correspond to machine instructions. Others like ENTER, LEAVE, RTN,
// ARG, CLL, RTV are used for managing calls amongst the components of
// an assembled program. There are also LBL instructions for labelling
// code sections for BRA, CLL, etc.
//
//
// Constructors for INST subclasses
// --------------------------------
//
// These do little work other than fill in the struct's info from the
// parameters.
//
// These are built during PASS 1 of compilation, when we traverse the
// syntax trees of each of the Cobra program's function definitions. 
//
//
// Methods of INST:
// ----------------
//
// There are only two methods supported by INSTs subclasses.
//
// * frameVars - This scans through the pseudo-instruction and assigns
//               a new slots within the assembled stack frame of the
//               function for anyly yet-allocated temporary variables
//               used by the instruction. This is performed in PASS 2
//               of compilation.
//
// * toMIPS - This converts the pseudo-instruction into a sequence of
//            MIPS instructions, outputting them to the give output
//            stream. This is performed in PASS 3 of compilation.
//
// Both of these methods take an Assembly object. This object tracks
// information for assembling each function component of the Cobra
// program. It also tracks whole-program information like string
// constants.
//
// The Assembly class, and its support class AssemblyComponent are
// defined and documented at the bottom of this file.
//
//

class INST {
public:
  virtual void frameVars(Assembly& assm) { }
  virtual void toMIPS(std::ostream& os, const Assembly& assm) const = 0;
};

typedef std::shared_ptr<INST> INST_ptr;
typedef std::vector<INST_ptr> INST_vec;

//
// Basic pseudo-instructions.
//
//   SET d,v     - sets a temporary to an integer value.
//   MOV d,s     - copies one temporary's value to another's.
//   ADD d,s1,s2 - sums two temporaries into another.
//   SUB, MUL    - similar; for subtraction and multiplication.
//   INC d,s     - adds a value to a temporary's value. 
//   NOP         - does nothing; "no operation"
//
class SET : public INST {
public:
  Name dst;
  int         val;
  SET(Name d, int v) : dst {d}, val {v} { }
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class MOV : public INST {
public:
  Name dst;
  Name src;
  MOV(Name d, Name s) : dst {d}, src {s} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class ADD : public INST {
public:
  Name dst;
  Name src1;
  Name src2;
  ADD(Name d, Name s1, Name s2) : dst {d}, src1 {s1}, src2 {s2} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class SUB : public INST {
public:
  Name dst;
  Name src1;
  Name src2;
  SUB(Name d, Name s1, Name s2) : src1 {s1}, src2 {s2}, dst {d} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class MUL : public INST {
public:
  Name dst;
  Name src1;
  Name src2;
  MUL(Name d, Name s1, Name s2) : src1 {s1}, src2 {s2}, dst {d} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class INC : public INST {
public:
  Name dst;
  Name src;
  INC(Name d, Name s) : dst {d}, src {s} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class NOP : public INST {
public:
  NOP(void) { } 
  void toMIPS(std::ostream& os, const Assembly& assm) const;
};


//
// Pseudo-instructions for jumping around within the code.
//
//   LBL l       - labels the top of a series of pseudo-instructions
//   BRA l       - branch unconditionally
//   BEQ s1,s2,l - branch if equal
//   BNE, BGT, BGE, BLT, BLE - similar; if greater than or equal
//
class LBL : public INST {
public:
  Label lbl;
  LBL(Label l) : lbl {l} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
};

class BCN : public INST {
public:
  std::string cndn; // One of "eq", "ne", "gt", "ge", "lt", "le", 
  Name src1;
  Name src2;
  Label lbl;
  BCN(std::string cn, std::string  s1, Name s2, Label l) :
    cndn {cn}, src1 {s1}, src2 {s2}, lbl {l} {}
  virtual void toMIPS(std::ostream& os, const Assembly& assm) const;
  virtual void frameVars(Assembly& assm);
};

class BEQ : public BCN {
public:
  BEQ(std::string  s1, Name s2, Label l) : BCN {"eq",s1,s2,l} { }
};

class BNE : public BCN {
public:
  BNE(std::string  s1, Name s2, Label l) : BCN {"ne",s1,s2,l} { }
};

class BLT : public BCN {
public:
  BLT(std::string  s1, Name s2, Label l) : BCN {"lt",s1,s2,l} { }
};

class BLE : public BCN {
public:
  BLE(std::string  s1, Name s2, Label l) : BCN {"le",s1,s2,l} { }
};

class BGT : public BCN {
public:
  BGT(std::string  s1, Name s2, Label l) : BCN {"gt",s1,s2,l} { }
};

class BGE : public BCN {
public:
  BGE(std::string  s1, Name s2, Label l) : BCN {"ge",s1,s2,l} { }
};

class BRA : public INST {
public:
  Label lbl;
  BRA(Label l) : lbl {l} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
};

//
// Pseudo-instructions for the called function.
//
// ENTER - sets up call stack frame for the function.
// RTN s - sets the return value
// LEAVE - takes down frame; returns
// 
//
class ENTER : public INST {
public:
  ENTER(void) {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
};

class RTN : public INST {
public:
  Name src;
  RTN(Name s) : src {s} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class LEAVE : public INST {
public:
  LEAVE(void) {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
};

//
// Pseudo-instructions for calling a function.
//
// ARG i,s - sets the i-th argument for a call
// CLL l   - calls the labelled function code
// RTV d   - gets the returned value
// 
//
class ARG : public INST {
public:
  int idx;
  Name src;
  ARG(int i, Name s) : idx {i}, src {s} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class RTV : public INST {
public:
  Name dst;
  RTV(Name d) : dst {d} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class CLL : public INST {
public:
  Label lbl;
  CLL(Label l) : lbl {l} {}
  void toMIPS(std::ostream& os, const Assembly& assm) const;
};

//
// Pseudo-instructions for system calls.
//
// RND d  - Gets a random value and puts it in a temporary variable.
// GET d  - Gets an input value and puts it in a temporary variable.
// PUTI s - Outputs the value of a temporary variable.
// PUTS l - Outputs a string at a label.
//
//
class RND : public INST {
public:
  Name dst;
  RND(Name d) : dst {d} {} 
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class GET : public INST {
public:
  Name dst;
  GET(Name d) : dst {d} {} 
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class PUTI : public INST {
public:
  Name src;
  PUTI(Name s) : src {s} { } 
  void toMIPS(std::ostream& os, const Assembly& assm) const;
  void frameVars(Assembly& assm);
};

class PUTS : public INST {
public:
  Label lbl;
  PUTS(Label l) : lbl {l} { } 
  void toMIPS(std::ostream& os, const Assembly& assm) const;
};


//
//
// ************************************************************


//
// class Assembly
// struct Assembly::Component
//
// An Assembly object is used for tracking the work of compiling an
// assembled Cobra program for converting it into MIPS code. Since a
// Cobra program can consist of several functions, including "main",
// it consists of a collection of assembly Component objects, one
// for each function definition.
//
// Each Component object tracks the assembled pseudo-instructions
// compiled for the execution of a function's body. It also maintains
// information about the stack frame built for that function's
// execution. Each variable used by the pseudo-instructions of the
// function's code has a designated slot within the function's
// stack frame. 
  
class Assembly {

private:

  class Component {
  public:
    Defn_ptr function; // Which function.
    std::unordered_map<Name,int> frame; // Slot location for each temporary.
    INST_vec instructions; // Assembled pseudo-instructions.
    Component(Defn_ptr fd) : function {fd}, instructions {}, frame {} {}
  };

  int lbl_counter;
  int tmp_counter;
  std::shared_ptr<Component> current;
  std::unordered_map<Name,std::shared_ptr<Component>> comps;
  std::vector<std::pair<Label,std::string>> strings;
  
public:
  Assembly(void);

  //
  // Management of assembled function subcomponents.
  //
  void addComponent(Defn_ptr fdef);
  Defn_ptr currentFunc(void) const;

  //
  // Support for assembling a program's string constants.
  //
  void dataString(Label lbl, std::string data);

  //
  // Support for PHASE 1 of compilation, converting the syntax trees
  // to pseudo-instructions.
  //
  std::string generateLabel(std::string baseName);
  std::string generateTmp(std::string baseName);
  std::string generateTmp(void);
  void append(INST_ptr inst);         

  //
  // Support for PHASE 2 of compilation, placing variables.
  //
  int frameSize(void) const;     // Get the whole frame size.  
  int offsetFor(Name var) const; // Get the slot location for `var`
  void frameVar(Name var);       // Find a slot for `var`.
  void frameVars(void);          // Find slots for all the code's variables.

  //
  // Support for PHASE 3 of compilation, generating the MIPS code.
  //
  void toMIPS(std::ostream& os);
};

#endif
