#include "cobra-ast.hh"
#include <string>
#include <vector>
#include <unordered_map>
#include <memory>
#include <utility>
#include <functional>
#include <iostream>
#include <exception>
#include <algorithm>

Prgm::Prgm(Defn_vec ds, Stmt_ptr mn) :
  defns {ds},
  // We wrap the main script into a function called "main".
  // Here we simply call that function.
  main {Stmt_ptr{new PCll{"main",Expn_vec{}}}}
{
  Defn_ptr d {new Defn{"main",Name_vec{},mn}};
  defns.push_back(d);
}

//
// global_ALL_DEFNS
//
// When a Cobra program executes, this dictionary gets populated with
// all the function definitions of the program so that procedure call
// statements and function call expressions can be handled. The code
// for pcll::exec accesses this dictionary, and the code for fcll::eval
// accesses this dictionary.
//
std::unordered_map<Name,Defn_ptr> Prgm::global_ALL_DEFNS { };

//
// rtrn_valu
//
// This defines a C++ exception to be thrown/caught so that we can
// properly handle exec::rtrn statements. We use C++ exceptions as
// the mechanism for getting that pcll::exec and fcll::eval to exit
// early, returning to their call site when `return` is handled 
// anywhere within a function's body.
//
class RtrnValu : public std::exception {
public:
  int rtvl;
  RtrnValu(int rv) : rtvl {rv} { }
};

//
// The code below gives helper functions for dealing with
// vectors of strings and vectors of `expn` pointers. There
// are probably cleaner/terser ways to do this.
//
// They are used in various `output` methods below.
//
std::string to_string(std::string bgn,
		      Name_vec nms,
		      std::string sep,
		      std::string end) {
  std::string s = bgn;
  bool first = true;
  for (Name nm : nms) {
    if (!first) {
      s += sep;
    }
    s += nm;
    first = false;
  }
  s += end;
  return s;
}
//
//
void voutput(std::ostream& os,
	     std::string bgn,
	     Expn_vec ev,
	     std::string sep,
	     std::string end) {
  os << bgn;
  bool first = true;
  for (Expn_ptr e : ev) {
    if (!first) {
      os << sep;
    }
    e->output(os);
    first = false;
  }
  os << end;
}


// **********
// class Defn
//

//
//  Defn::output(os)
//
//  Sends formatted code for a Cobra function definition
//  to the given output stream `os`.
//
void Defn::output(std::ostream& os) const {
  os << "def ";
  os << name;
  os << to_string("(",prams," ","):");
  os << std::endl;
  body->output(os,"  ");
  os << std::endl;
}

//
// Defn::call(vs)
//
// Calls the Cobra function definition with parameters
// given by the vector `vs`.
//
// NOTE: assumes the size of `vs` is the size of `this->prams`.
//
// This involves the following:
//
//  1. Set up an empty frame, then binds each parameter
//     variable to the corresponding argument value,
//     storing each in the frame.
//  2. Execute the body of the function.
//  3. During #2, catch any thrown "return value" exceptions.
//
int Defn::call(const std::vector<int>& vs) const {
  //
  // Create a fresh call frame.
  Frme frme { };
  //
  // Add all the passed parameters to the frame.
  auto vp = vs.begin();
  for (Name p : prams){
    if (vp != vs.end()) {
      frme[p] = (*vp);
    } else {
      frme[p] = 0;
    }
    vp++;
  }
  try {
    //
    // Execute the body.
    body->exec(frme);
    return 0;
  } catch (RtrnValu rv) {
    //
    // Catch any returned values.
    return rv.rtvl;
  }
}
	       
void Asgn::output(std::ostream& os, std::string indent) const {
  os << indent;
  os << name << " = ";
  valu->output(os);
  os << std::endl;
}

void Asgn::exec(Frme& frme) const {
  frme[name] = valu->eval(frme);
}

void Updt::output(std::ostream& os, std::string indent) const {
  os << indent;
  os << name << " += ";
  valu->output(os);
  os << std::endl;
}

void Updt::exec(Frme& frme) const {
  /* to do for Exercise 5.2 */
}

void Rtrn::output(std::ostream& os, std::string indent) const {
  os << indent << "return ";
  valu->output(os);
  os << std::endl;
}

void Rtrn::exec(Frme& frme) const {
  throw RtrnValu {valu->eval(frme)};
}

void Pass::output(std::ostream& os, std::string indent) const {
  os << indent << "pass" << std::endl;
}

void Pass::exec(Frme& frme) const {
  // does nothing!
}

void ForL::output(std::ostream& os, std::string indent) const {
  os << indent;
  os << "for " << name << " = " << frst;
  os << " to " << last << ":" << std::endl;
  body->output(os,indent+"  ");
}

void ForL::exec(Frme& frme) const {
  /* to do for Exercise 5.6 */
}

void Whle::output(std::ostream& os, std::string indent) const {
  os << indent;
  os << "while";
  test->output(os);
  os << ":" << std::endl;
  body->output(os,indent+"  ");
}

void Whle::exec(Frme& frme) const {
  while (test->chck(frme)) {
    body->exec(frme);
  } 
}

void IfTn::output(std::ostream& os, std::string indent) const {
  os << indent;
  os << "if";
  test->output(os);
  os << ":" << std::endl;
  tbdy->output(os,indent+"  ");
}

void IfTn::exec(Frme& frme) const {
  /* to do for Exercise 5.1 */
}

void IfTE::output(std::ostream& os, std::string indent) const {
  os << indent;
  os << "if";
  test->output(os);
  os << ":" << std::endl;
  tbdy->output(os,indent+"  ");
  os << "else:" << std::endl;
  ebdy->output(os,indent+"  ");
}

void IfTE::exec(Frme& frme) const {
  if (test->chck(frme)) {
    tbdy->exec(frme);
  } else {
    ebdy->exec(frme);
  }
}

void PCll::output(std::ostream& os, std::string indent) const {
  os << indent;
  os << name;
  voutput(os,"(",argms,",",")");
  os << std::endl;
}

void PCll::exec(Frme& frme) const {
  std::vector<int> valus { };
  valus.resize(argms.size());
  std::function<int(Expn_ptr)> eval_arg = [&frme](Expn_ptr a) -> int {
					    return a->eval(frme);
					  };
  std::transform(argms.begin(),argms.end(),valus.begin(),eval_arg);
  Prgm::global_ALL_DEFNS[name]->call(valus);
}
  
void PrtS::output(std::ostream& os, std::string indent) const {
  os << indent;
  os << "print";
  os << "(\"";
  os << outs;
  os << "\")";
  os << std::endl;
}

void PrtS::exec(Frme& frme) const {
  std::cout << outs << std::endl;
}

void PrtI::output(std::ostream& os, std::string indent) const {
  os << indent;
  os << "print";
  os << "(";
  valu->output(os);
  os << ")";
  os << std::endl;
}

void PrtI::exec(Frme& frme) const {
  std::cout << valu->eval(frme) << std::endl;
}

void Blck::output(std::ostream& os, std::string indent) const {
  os << indent << "begin" << std::endl;
  for (Stmt_ptr s : stmts) {
    s->output(os,indent+"  ");
  }
  os << indent << "end" << std::endl;
}

void Blck::exec(Frme& frme) const {
  for (Stmt_ptr s : stmts) {
    s->exec(frme);
  }
}

void Plus::output(std::ostream& os) const {
  os << "(";
  left->output(os);
  os << " + ";
  rght->output(os);
  os << ")";
}

int Plus::eval(const Frme& frme) const {
  int lv = left->eval(frme);
  int rv = rght->eval(frme);
  return (lv + rv);
}

void Mnus::output(std::ostream& os) const {
  os << "(";
  left->output(os);
  os << " - ";
  rght->output(os);
  os << ")";
}

int Mnus::eval(const Frme& frme) const {
  /* to do for Exercise 5.3 */
  return 0;
}

void Tmes::output(std::ostream& os) const {
  os << "(";
  left->output(os);
  os << " * ";
  rght->output(os);
  os << ")";
}

int Tmes::eval(const Frme& frme) const {
  int lv = left->eval(frme);
  int rv = rght->eval(frme);
  return (lv * rv);
}

void AndC::output(std::ostream& os) const {
  os << "(";
  left->output(os);
  os << " and ";
  rght->output(os);
  os << ")";
}

bool AndC::chck(const Frme& frme) const {
  return left->chck(frme) && rght->chck(frme);
}

void OrCn::output(std::ostream& os) const {
  os << "(";
  left->output(os);
  os << " or ";
  rght->output(os);
  os << ")";
}

bool OrCn::chck(const Frme& frme) const {
  /* to do for Exercise 5.5 */
  return true;
}

void NotC::output(std::ostream& os) const {
  os << "(not ";
  cndn->output(os);
  os << ")";
}

bool NotC::chck(const Frme& frme) const {
  return !cndn->chck(frme);
}

void Less::output(std::ostream& os) const {
  os << "(";
  left->output(os);
  os << " < ";
  rght->output(os);
  os << ")";
}

bool Less::chck(const Frme& frme) const {
  int lv = left->eval(frme);
  int rv = rght->eval(frme);
  return (lv < rv);
}

void Eqls::output(std::ostream& os) const {
  os << "(";
  left->output(os);
  os << " < ";
  rght->output(os);
  os << ")";
}

bool Eqls::chck(const Frme& frme) const {
  int lv = left->eval(frme);
  int rv = rght->eval(frme);
  return (lv == rv);
}

void LEql::output(std::ostream& os) const {
  os << "(";
  left->output(os);
  os << " <= ";
  rght->output(os);
  os << ")";
}

bool LEql::chck(const Frme& frme) const {
  /* to do for Exercise 5.4 */
  return true;
}

void Nmbr::output(std::ostream& os) const {
  os << std::to_string(valu);
}

int Nmbr::eval(const Frme& frme) const {
  return valu;
}

void Lkup::output(std::ostream& os) const {
  os << name;
}

int Lkup::eval(const Frme& frme) const {
  return frme.at(name);
}

void Rndm::output(std::ostream& os) const {
  os << "random(";
  maxv->output(os);
  os << ")";
}

int Rndm::eval(const Frme& frme) const {
  return (rand() % maxv->eval(frme));
}

void Inpt::output(std::ostream& os) const {
  os << "input(\"" << prpt << "\")";
}

int Inpt::eval(const Frme& frme) const {
  int vl;
  std::cout << prpt;
  std::cin >> vl;
  return vl;
}

void FCll::output(std::ostream& os) const {
  os << name;
  voutput(os,"(",argms,",",")");
}

int FCll::eval(const Frme& frme) const {
  std::vector<int> valus { };
  valus.resize(argms.size());
  std::function<int(Expn_ptr)> eval_arg = [&frme](Expn_ptr a) -> int {
					    return a->eval(frme);
					  };
  std::transform(argms.begin(),argms.end(),valus.begin(),eval_arg);
  return Prgm::global_ALL_DEFNS[name]->call(valus);
}

void Prgm::output(std::ostream& os) const {
  for (Defn_ptr d : defns) {
    d->output(os);
    os << std::endl;
  }
  main->output(os);
}

void Prgm::run(void) const {
  srand(time(0));
  for (Defn_ptr d : defns) {
    Prgm::global_ALL_DEFNS[d->name] = d;
  }
  Frme main_frme { };
  main->exec(main_frme);
}
