#ifndef __COBRA_AST_H_
#define __COBRA_AST_H_

//
// Object classes used by the parser to represent the syntax trees of
// Cobra programs.
//
// It defines two top-level classes for the Cobra program and its
// function definition, and then three major abstract classes whose
// subclasses form the syntax trees of a function's code.
//
//  * Prgm - a Cobra program; made up of several function definitions
//           including "main".
//
//  * Defn - a Cobra function definition; its parameters and body
//
//  * Stmt - subclasses for the various statments that can occur
//           in a function body: loops, conditionals, updates, etc.
//           These get executed when the program runs.
//
//  * Expn - subclasses for the various integer-valued expressions
//           than can occur on the right-hand side of an assignment
//           statement. These get evaluated to compute a value.
//
//  * Cndn - subclasses for the various boolean-valued conditions
//           that govern the execution of a loop or an "if". These
//           get checked to determine the "flow of control" through
//           the statements of the program.
//
//

#include <string>
#include <vector>
#include <unordered_map>
#include <memory>
#include <utility>
#include <iostream>

// Needed to allow for mutually recursive definitions of classes.
class PCll;
class Prgm;
class Defn;
class Stmt;
class Expn;
class Cndn;

#ifdef COBRAC
class Assembly;
#endif

//
// Below are "type (name) definitions". They allow us to invent new
// names for types of data. Here they will help us have shorter names
// for all the types we use below.
//
typedef std::string Label;
typedef std::string Name;
typedef std::vector<Name>     Name_vec;
typedef std::shared_ptr<Prgm> Prgm_ptr; 
typedef std::shared_ptr<Defn> Defn_ptr; 
typedef std::shared_ptr<Stmt> Stmt_ptr; 
typedef std::shared_ptr<Expn> Expn_ptr; 
typedef std::shared_ptr<Cndn> Cndn_ptr; 
typedef std::vector<Defn_ptr> Defn_vec;
typedef std::vector<Stmt_ptr> Stmt_vec;
typedef std::vector<Expn_ptr> Expn_vec;
typedef std::unordered_map<Name,int> Frme;

//
// class Prgm
//
// An object in this class holds all the information gained
// from parsing the source code of a Cobra program. A program
// is a series of Cobra function definitions followed by the
// (main) script of statements that execute that program using
// those functions.
//
// The current working assumption of the code is that there is
// only one Prgm object instance when a client uses this code.
//
// Two kinds of client are supported:
//  		     
//  * a Cobra interpreter - this simply runs the Cobra program,
//        executing its statments, updating the state of program
//        variables as a result, getting user input from the
//        console, and outputting results to the console.
//
//  * a Cobra to MIPS compiler - this assembles the main and
//        auxilliary functions of a Cobra program into MIPS
//        code that can be run with the SPIM interpreter.
//
// For the interpreter, the Prgm::run method is used. It relies
// on use of the class variable `global_ALL_DEFS` to provide "global"
// access by the client to the Cobra program's suite of function
// definitions. The interpreter relies on the Defn::call, Stmt::exec,
// Expn::eval, and Cndn::chck methods of the various syntactic
// components that constitute the Prgm object and its Defn objects.
//
// For the compiler, the Prgm::compile method is used. It relies on a
// separate object type called the `Assembly` that serves as the
// intermeidate data structure for converting the Cobra source code to
// the compiled MIPS code. Compilation relies on the Defn::assemble,
// Stmt::assemble, Expn::assembleToSet, and Cndn::assembleAsBranch
// methods.
//
//
		     
class Prgm {
public:
  //
  static std::unordered_map<Name,Defn_ptr> global_ALL_DEFNS;
  //
  Defn_vec defns;
  Stmt_ptr main;
  //
  Prgm(Defn_vec ds, Stmt_ptr mn);
  //
  void run(void) const; // Execute the program, interpreting its code.
#ifdef COBRAC
  void compile(std::ostream& os) const; // Convert the program to MIPS.
  void assemble(Assembly& code) const;     // Build an intermediate rep'n.
#endif
  void output(std::ostream& os) const;  // Output formatted code.
};


//
// class Prgm
//
// An object in this class holds all the information gained
// from parsing the source code of a Cobra program. A program
class Defn {
public:
  //
  Name     name;  // the name of the function  
  Name_vec prams; // the names of its parameters
  Stmt_ptr body;  // the statement(s) that form its body
  //
  Defn(Name nm, Name_vec ps, Stmt_ptr bd) :
    name {nm}, prams {ps}, body {bd} { }
  //
  int call(const std::vector<int>& vs) const; // Run with arguments.
#ifdef COBRAC
  void assemble (Assembly& code) const;   // Build its intermediate rep'n.
#endif
  void output(std::ostream& os) const; // Output its code, formatted. 
};

//
//
// ************************************************************

// Syntax tree classes used to represent a Cobra program's code.
//
// classes Stmt, Expn, Chck, all subclasses of AST.
//

//
// class AST 
//
// Cover type for all "abstract" syntax tree classes.
//

class AST {
public:
  virtual void output(std::ostream& os) const = 0;
};

//
// class Stmt
//
// Abstract class for program statment syntax trees,
//
// Subclasses are
//
//   Asgn - assignment statement "v = e"
//   Updt - increment statement "v += e"
//   Whle - while loop statememt "while c: s"
//   ForL - for loop statememt "for v = e1 to e2: s"
//   IfTn - conditional statement "if c: s"
//   IfTE - conditional statement "if c: s1 else: s2"
//   PrtI - value output statement "print(e)"
//   PrtS - string output statement "print("some text here")"
//   Call - procedure call statement "f(e1,e2,...,ek)"
//   Rtrn - return statement "return e"
//   Pass - statement that does nothing
//   Blck - "compound statement for a block of statements
//           "begin s1 s2 ... sk end"
//
// These each support the methods:
//
//  * exec(frme): execute the statement within the stack frame
//
//  * assemble(assm): convert the statement to a (series of) pseudo
//        instructions. Those get appended to the `assm` object.
//
//  * output(os), output(os,indent): output formatted Cobra code of
//        the statement to the output stream `os`. The `indent` string
//        gives us a string of spaces for indenting the lines of its
//        code.
//
//
class Stmt : public AST {
public:
  virtual void exec(Frme& frme) const = 0;
#ifdef COBRAC
  virtual void assemble(Assembly& code) const = 0;
#endif
  virtual void output(std::ostream& os, std::string indent) const = 0;
  virtual void output(std::ostream& os) const {
    output(os,"");
  }
};

class Asgn : public Stmt {
public:
  Name     name;
  Expn_ptr valu;
  Asgn(Name nm, Expn_ptr vl) : name {nm}, valu {vl} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class Updt : public Stmt {
public:
  Name     name;
  Expn_ptr valu;
  Updt(Name nm, Expn_ptr vl) : name {nm}, valu {vl} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class PrtI : public Stmt {
public:
  Expn_ptr valu;
  PrtI(Expn_ptr vl) : valu {vl} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class PrtS : public Stmt {
public:
  std::string outs;
  PrtS(std::string out) : outs {out} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class Rtrn : public Stmt {
public:
  Expn_ptr valu;
  Rtrn(Expn_ptr vl) : valu {vl} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class Pass : public Stmt {
public:
  Pass(void) { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class ForL : public Stmt {
public:
  Name     name;
  Expn_ptr frst;
  Expn_ptr last;
  Stmt_ptr body;
  ForL(Name nm, Expn_ptr ft, Expn_ptr lt, Stmt_ptr bd) :
    name {nm}, frst {ft}, last {lt}, body {bd} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class Whle : public Stmt {
public:
  Cndn_ptr test;
  Stmt_ptr body;
  Whle(Cndn_ptr cn, Stmt_ptr bd) : test {cn}, body {bd} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class IfTn : public Stmt {
public:
  Cndn_ptr test;
  Stmt_ptr tbdy;
  IfTn(Cndn_ptr cn, Stmt_ptr tb) : test {cn}, tbdy {tb} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class IfTE : public Stmt {
public:
  Cndn_ptr test;
  Stmt_ptr tbdy;
  Stmt_ptr ebdy;
  IfTE(Cndn_ptr cn, Stmt_ptr tb, Stmt_ptr eb) :
    test {cn}, tbdy {tb}, ebdy {eb} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class Blck : public Stmt {
public:
  Stmt_vec stmts;
  Blck(Stmt_vec ss) : stmts {ss} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};

class PCll : public Stmt {
public:
  Name     name;
  Expn_vec argms;
  PCll(Name nm, Expn_vec as) : name {nm}, argms {as} { }
  void exec(Frme& frme) const;
#ifdef COBRAC
  void assemble(Assembly& code) const;
#endif
  void output(std::ostream& os, std::string indent) const;
};


//
// class Expn
//
// Abstract class for integer expression syntax trees,
//
// Subclasses are
//
//   Plus - binary operation + applied to two sub-expressions
//   Mnus - binary operation - applied to two sub-expressions
//   Tmes - binary operation * applied to two sub-expressions
//   Nmbr - integer constant expression
//   Lkup - variable access (i.e. "look-up") within function frame
//   Inpt - obtains an integer input (after output of a prompt)
//   Rndm - yields a random value
//   FCll - apply a function to some subexpressions
//
// These each support the methods:
//
//  * eval(frme): evaluate the expression; return its result
//
//  * assembleToSet(dst, assm): convert the statement to a (series of)
//        pseudo-instructions. Those get appended to the `assm`
//        object.  The instructions should ultimately place the
//        expressions value into the `dst` temporary variable.
//
//  * output(os): output formatted Cobra code of the expression.
//
//
class Expn : public AST {
public:
  virtual  int eval(const Frme& frme) const = 0;
#ifdef COBRAC
  virtual void assembleToSet(Name dst, Assembly& code) const = 0;
#endif
};

class Plus : public Expn {
public:
  Expn_ptr left;
  Expn_ptr rght;
  Plus(Expn_ptr lf, Expn_ptr rg) : left {lf}, rght {rg} { }
  int eval(const Frme& frme) const;
#ifdef COBRAC
  void assembleToSet(Name dst, Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class Mnus : public Expn {
public:
  Expn_ptr left;
  Expn_ptr rght;
  Mnus(Expn_ptr lf, Expn_ptr rg) : left {lf}, rght {rg} { }
  int eval(const Frme& frme) const;
#ifdef COBRAC
  void assembleToSet(Name dst, Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class Tmes : public Expn {
public:
  Expn_ptr left;
  Expn_ptr rght;
  Tmes(Expn_ptr lf, Expn_ptr rg) : left {lf}, rght {rg} { }
  int eval(const Frme& frme) const;
#ifdef COBRAC
  void assembleToSet(Name dst, Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class Nmbr : public Expn {
public:
  int valu;
  Nmbr(int vl) : valu {vl} { }
  int eval(const Frme& frme) const;
#ifdef COBRAC
  void assembleToSet(Name dst, Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class FCll : public Expn {
public:
  Name name;
  Expn_vec argms;
  FCll(Name nm, std::vector<Expn_ptr> as) : name {nm}, argms {as} { }
  int eval(const Frme& frme) const;
#ifdef COBRAC
  void assembleToSet(Name dst, Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};
  
class Lkup : public Expn {
public:
  Name name;
  Lkup(Name nm) : name {nm} { }
  int eval(const Frme& frme) const;
#ifdef COBRAC
  void assembleToSet(Name dst, Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class Inpt : public Expn {
public:
  std::string prpt;
  Inpt(std::string pr) : prpt {pr} { }
  int eval(const Frme& frme) const;
#ifdef COBRAC
  void assembleToSet(Name dst, Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};
  
class Rndm : public Expn {
public:
  Expn_ptr maxv;
  Rndm(Expn_ptr mx) : maxv {mx} { }
  int eval(const Frme& frme) const;
#ifdef COBRAC
  void assembleToSet(Name dst, Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};


//
// class Cndn
//
// Abstract class for boolean conditions.
//
// Its subclasses are
//   
//   NotC - conditional negation
//   AndC - conditional conjunction
//   OrCn - conditional disjunction
//   Less - conditional expression "e1 < e2"
//   Eqls - conditional expression "e1 == e2"
//   LEql - conditional expression "e1 <= e2"
//
// These each support the methods:
//
//  * chck(frme): execute the statement within the stack frame.
//
//  * assembleToBranchIfTrue(tlbl,assm): convert the statement to a
//        (series of) pseudo-instructions. Those get appended to the
//        `assm` object. The instructions will branch to the given
//        label when the condition is determined to be true.
//
//  * assembleToBranchIfFalse(flbl,assm): convert the statement to a
//        (series of) pseudo-instructions. Those get appended to the
//        `assm` object. The instructions will branch to the given
//        label when the condition is determined to be false.
//
//  * output(os): output formatted Cobra code of the condition.
//
//
class Cndn : public AST {
public:
  virtual bool chck(const Frme& frme) const = 0;
#ifdef COBRAC
  virtual void assembleToBranchIfTrue(Label tlbl,Assembly& code) const = 0;
  virtual void assembleToBranchIfFalse(Label flbl,Assembly& code) const = 0;
#endif
};

class AndC : public Cndn {
public:
  Cndn_ptr left;
  Cndn_ptr rght;
  AndC(Cndn_ptr lf, Cndn_ptr rg) : left {lf}, rght {rg} { }
  bool chck(const Frme& frme) const;
#ifdef COBRAC
  void assembleToBranchIfTrue(Label tlbl,Assembly& code) const;
  void assembleToBranchIfFalse(Label flbl,Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class OrCn : public Cndn {
public:
  Cndn_ptr left;
  Cndn_ptr rght;
  OrCn(Cndn_ptr lf, Cndn_ptr rg) : left {lf}, rght {rg} { }
  bool chck(const Frme& frme) const;
#ifdef COBRAC
  void assembleToBranchIfTrue(Label tlbl,Assembly& code) const;
  void assembleToBranchIfFalse(Label flbl,Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class NotC : public Cndn {
public:
  Cndn_ptr cndn;
  NotC(Cndn_ptr cn) : cndn {cn} { }
  bool chck(const Frme& frme) const;
#ifdef COBRAC
  void assembleToBranchIfTrue(Label tlbl,Assembly& code) const;
  void assembleToBranchIfFalse(Label flbl,Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class Less : public Cndn {
public:
  Expn_ptr left;
  Expn_ptr rght;
  Less(Expn_ptr lf, Expn_ptr rg) : left {lf}, rght {rg} { }
  bool chck(const Frme& frme) const;
#ifdef COBRAC
  void assembleToBranchIfTrue(Label tlbl,Assembly& code) const;
  void assembleToBranchIfFalse(Label flbl,Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class Eqls : public Cndn {
public:
  Expn_ptr left;
  Expn_ptr rght;
  Eqls(Expn_ptr lf, Expn_ptr rg) : left {lf}, rght {rg} { }
  bool chck(const Frme& frme) const;
#ifdef COBRAC
  void assembleToBranchIfTrue(Label tlbl,Assembly& code) const;
  void assembleToBranchIfFalse(Label flbl,Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};

class LEql : public Cndn {
public:
  Expn_ptr left;
  Expn_ptr rght;
  LEql(Expn_ptr lf, Expn_ptr rg) : left {lf}, rght {rg} { }
  bool chck(const Frme& frme) const;
#ifdef COBRAC
  void assembleToBranchIfTrue(Label tlbl,Assembly& code) const;
  void assembleToBranchIfFalse(Label flbl,Assembly& code) const;
#endif
  void output(std::ostream& os) const;
};


#endif
