%skeleton "lalr1.cc" /* -*- C++ -*- */
%require "3.0.4"

%defines
%define api.parser.class {cobra_parser}
%define api.token.constructor
%define api.value.type variant
%define parse.assert

%code requires
{
#include "cobra-ast.hh"
#include <string>
#include <vector>
#include <unordered_map>
#include <memory>
#include <utility>
class cobra;
}
// The parsing context.
%param { cobra& driver }
%locations
%initial-action
{
  // Initialize the initial location.
  @$.begin.filename = @$.end.filename = &driver.file;
};
%define parse.trace
%define parse.error verbose
%code
{
#include "cobra-ast.hh"
#include <string>
#include <vector>
#include <unordered_map>
#include <memory>
#include <utility>
#include "cobra.hh"
}
%define api.token.prefix {TOK_}
%token
  EOF  0  "end of file"
  DEF     "def"
  BEGIN   "begin"
  END     "end"
  WHILE   "while"
  FOR     "for"
  TO      "to"
  IF      "if"
  ELSE    "else"
  PASS    "pass"
  PRINT   "print"
  INPUT   "input"
  RANDOM  "random"
  RETURN  "return"
  ASSIGN  "="
  UPDATE  "+="
  PLUS    "+"
  MINUS   "-"
  TIMES   "*"
  COLON   ":"
  COMMA   ","
  LPAREN  "("
  RPAREN  ")"
  LESS    "<"
  EQUALS  "=="
  LEQUAL  "<="
  AND     "and"
  OR      "or"
  NOT     "not"
;

%token <std::string> NAME
%token <int>         NMBR
%token <std::string> STRG

%type <Prgm_ptr> prgm
%type <Defn_ptr> defn
%type <Stmt_ptr> stmt
%type <Expn_ptr> expn
%type <Cndn_ptr> cndn

%type <Name_vec> names
%type <Defn_vec> defns
%type <Stmt_vec> stmts
%type <Expn_vec> expns

%%

%start prgm;

%precedence COLON;
%precedence ELSE;

%left OR;
%left AND;
%precedence NOT;
%left PLUS MINUS;
%left TIMES;

prgm: defns stmt {
  Prgm_ptr p = Prgm_ptr{new Prgm {$1,$2}};
  driver.result = p;
  $$ = p;
 }
;

defns
: %empty { }
| defns defn {
  Defn_vec ds = $1;
  Defn_ptr d = $2;
  ds.push_back(d);
  $$ = ds;
 }
;

defn
: DEF NAME LPAREN names RPAREN COLON stmt
                            { $$ = Defn_ptr {new Defn {$2,$4,$7}}; }
;

stmt
: NAME ASSIGN expn          { $$ = Stmt_ptr {new Asgn {$1,$3}}; }
| NAME UPDATE expn          { $$ = Stmt_ptr {new Updt {$1,$3}}; }
| NAME LPAREN expns RPAREN  { $$ = Stmt_ptr {new PCll {$1,$3}}; }
| PRINT LPAREN STRG RPAREN  { $$ = Stmt_ptr {new PrtS {$3}}; }
| PRINT LPAREN expn RPAREN  { $$ = Stmt_ptr {new PrtI {$3}}; }
| FOR NAME ASSIGN expn TO expn COLON stmt
                            { $$ = Stmt_ptr {new ForL {$2,$4,$6,$8}};} 
| WHILE cndn COLON stmt     { $$ = Stmt_ptr {new Whle {$2,$4}}; }
| IF cndn COLON stmt        { $$ = Stmt_ptr {new IfTn {$2,$4}}; /* to do */ } 
| IF cndn COLON stmt ELSE COLON stmt
                            { $$ = Stmt_ptr {new IfTE {$2,$4,$7}}; }
| RETURN expn               { $$ = Stmt_ptr {new Rtrn {$2}}; } 
| PASS                      { $$ = Stmt_ptr {new Pass {}}; }
| BEGIN stmts END           { $$ = Stmt_ptr {new Blck {$2}}; }
;

cndn
: expn LESS expn            { $$ = Cndn_ptr {new Less {$1,$3}}; }
| expn EQUALS expn          { $$ = Cndn_ptr {new Eqls {$1,$3}}; }
| expn LEQUAL expn          { $$ = Cndn_ptr {new LEql {$1,$3}}; /* to do */ }
| cndn AND cndn             { $$ = Cndn_ptr {new AndC {$1,$3}}; /* to do */ }
| cndn OR cndn              { $$ = Cndn_ptr {new OrCn {$1,$3}}; /* to do */ }
| NOT cndn                  { $$ = Cndn_ptr {new NotC {$2}};    /* to do */ }
| LPAREN cndn RPAREN        { $$ = $2; }
;

expn
: expn PLUS expn            { $$ = Expn_ptr {new Plus {$1,$3}}; }
| expn MINUS expn           { $$ = Expn_ptr {new Mnus {$1,$3}}; /* to do */ }
| expn TIMES expn           { $$ = Expn_ptr {new Tmes {$1,$3}}; }
| RANDOM LPAREN expn RPAREN { $$ = Expn_ptr {new Rndm {$3}}; }
| INPUT LPAREN STRG RPAREN  { $$ = Expn_ptr {new Inpt {$3}}; }
| NAME LPAREN expns RPAREN  { $$ = Expn_ptr {new FCll {$1,$3}}; }
| NMBR                      { $$ = Expn_ptr {new Nmbr {$1}}; }
| NAME                      { $$ = Expn_ptr {new Lkup {$1}}; }
| LPAREN expn RPAREN        { $$ = $2; }
;

expns
: expn                      { $$ = Expn_vec { $1 }; }
| expns COMMA expn          { Expn_vec es = $1; es.push_back($3); $$ = es; }
;

stmts
: stmt                      { $$ = Stmt_vec { $1 }; }
| stmts stmt                { Stmt_vec ss = $1; ss.push_back($2); $$ = ss; }
;

names
: NAME                      { $$ = Name_vec { $1 }; }
| names COMMA NAME          { Name_vec ps = $1; ps.push_back($3); $$ = ps; }
;

%%
void yy::cobra_parser::error (const location_type& l,
			      const std::string& m) {
  driver.error (l, m);
}
