#include <memory>
#include <iostream>
#include <fstream>
#include <string>
#include "cobra-assem.hh"
#include "cobra.hh"

Assembly::Assembly(void) :
  lbl_counter {0},
  tmp_counter {0},
  comps { },
  strings { }
{ }

void Assembly::addComponent(Defn_ptr fdef) {
  current = std::shared_ptr<Component> {new Component{fdef}};
  comps[fdef->name] = current;
}

Defn_ptr Assembly::currentFunc(void) const {
  return current->function;
}

void Assembly::dataString(Label lbl, std::string data) {
  strings.push_back(std::pair<Label,std::string> {lbl,data});
}

void Assembly::append(std::shared_ptr<INST> inst) {
  current->instructions.push_back(inst);
}

std::string Assembly::generateLabel(std::string baseName) {
  std::string number = std::to_string(lbl_counter++);
  return current->function->name + "_" + baseName + "_" + number;
}

std::string Assembly::generateTmp(std::string baseName) {
  std::string number = std::to_string(tmp_counter++);
  return current->function->name + "_" + baseName + "_" + number;
}

std::string Assembly::generateTmp(void) {
  return generateTmp("tmp");
}

int Assembly::frameSize(void) const {
  int slots = current->frame.size();         
  int size = 4 * (slots + slots%2); // make it double-word aligned
  return size;
}

int Assembly::offsetFor(Name var) const {
  auto vp = current->frame.find(var);
  if (vp == current->frame.end()) {
    std::cout << "cobrac: While compiling function '" << current->function->name << "'..." << std::endl;
    std::cout << "\tBad variable '" << var << "'." <<std::endl;
  }
  return -(current->frame.at(var))*4;
}

void Assembly::frameVar(Name var) {
  if (current->frame.find(var) == current->frame.end()) {
    current->frame[var] = current->frame.size()+1;
  }
}

void Assembly::frameVars(void) {
  for (std::pair<Name,std::shared_ptr<Component>> fc : comps) {
    current = fc.second;
    frameVar("saved_return_address");
    frameVar("saved_frame_pointer");
    for (Name p : current->function->prams) {
      frameVar(p);
    }
    for (INST_ptr inst : current->instructions) {
      inst->frameVars(*this);
    }
  }
}

void Assembly::toMIPS(std::ostream& os) {
  os << "\t.data" << std::endl;
  for (auto ls : strings) {
    os << ls.first << ":\t.asciiz \"" << ls.second << "\"" << std::endl; 
  }
  os << "\t.globl main" << std::endl;
  os << "\t.text" << std::endl;
  for (std::pair<Name,std::shared_ptr<Component>> fc : comps) {
    current = fc.second;
    for (INST_ptr inst : current->instructions) { 
      inst->toMIPS(os,*this);
    }
  }
}




int main (int argc, char *argv[]) {

  std::string filename {argv[1]};
  std::string root = filename.substr(0,filename.length()-4);
  std::string extension = filename.substr(filename.length()-4);
  if (extension != ".cob") {
    std::cerr << "Not a cobra file. Exiting!" << std::endl;
  }

  cobra driver { };
  int error = driver.parse(filename);
  if (error) {
    std::cerr << "Too many errors to compile. Exiting!" << std::endl;
  }
  
  std::ofstream asmFile;
  asmFile.open(root+".s");
  driver.result->compile(asmFile);
  asmFile.close();

  return 0;
}

