#include "Gate.hh"
#include "Circuit.hh"
#include <fstream>
#include <iostream>
#include <vector>
#include <memory>
#include <map>
#include <algorithm>

//
// class Circuit
//
// A circuit is a collection of circuit components, each an instance
// of a subclass of Gate. It is constructed by providing a file name,
// one of a text file that contains the description of its components,
// along with how they should be wired together.
//
// A few sample circuit files are included in this folder.
//
//    xor.lsc: a two input XOR circuit
//
//    mux4.lsc: a 6 input circuit that behaves like a 4:1 MUX
//
//    inc.lsc: next state logic for a 3-bit cyclic counter
//
//    dec2_4: a 2:4 decoder circuit
//
// The Circuit components are defined in Gate.cc and Gate.hh.
//
// There are Input components. These are ones that can be set to
// true/false with the method setValue. Then there are And, Or, and
// Not gates. These take their input(s) from other conponents, and
// compute an output value based on their logical behavior. Finally,
// there are Output components, which simply reflect the value of
// their one input.
//
// There are four major methods for computing with a circuit.
//
//    reset: called before we want to compute the outputs
//           of a circuit from its inputs
//
//    setInputsFrom: called before we start that computation.
//           It sets each of the Input components to true/false
//           based on a vector<bool> with inputSize() elements.
//
//    evaluate: called to compute the outputs of a circuit from
//           its set inputs.
//
//    readOutputsInto: called with a vector<bool>. The truth
//           values of each of the outputs are read into this
//           vector.
//
// To compute with a Circuit `c`, follow these four steps.
//
//     1. Call `c.reset()`.
//     2. Call `c.setInputsFrom(ibv)` with some vector `ibv`.
//        Make sure its length is `c.inputSize()`.  This gets
//        the inputs ready.
//     3. Call `c.evaluate()`. This should simulate the 
//        circuit and ultimately set the outputs.
//     4. Call `c.readOutputsInto(obv)` with some vector `obv`.
//        Make sure its length is `c.outputSize()`. This reads
//        the outputs into that vector.
//
// The method `evaluate` is not written. That is part of the
// assignment.
//
// One test of `evaluate` is the method `outputTruthTable`.  This code
// generates a series of text lines that give the behavior of the
// circuit with all possible input patterns. It runs steps 1-4 on each
// input possible vector, and output the result.
// 
// See the `.out` files in this folder for the expected output of the
// truth table of each of the sample circuits.
//
// The method `outputTruthTable` is written similarly to the work you
// did for `Odometer` in Lab 13. Using a loop, it does the work of
// setting the inputs to some input pattern, evaluating the circuit
// on that pattern, and then reporting the inputs/outputs to std::cout
// as a row in a truth table. It loops through all the input patterns
// from "all set to false" to "all set to true".
//

// constructor
//
// Builds a Circuit object by processing an input text file.
//
Circuit::Circuit(std::string filename) :
  inputs {},
  outputs {},
  components {} 
{
  std::ifstream fs;

  // Open the file.
  //
  fs.open(filename);

  if (fs) {
    // Read it if no errors occurred.
    //
    std::cerr << "Reading '" << filename << "'...\n";
    _readFrom(fs);
  }

  // Close it.
  //
  fs.close();
}

// _readFrom
//
// Does the bulk of the work of the constructor.
//
void Circuit::_readFrom(std::ifstream& fs) {

  // Read the number of gate components.
  int numGates;
  fs >> numGates;

  // Read each gate component entry, add it to the circuit.
  //
  // This involves several steps:
  //  1. create a shared_ptr to the new gate.
  //  2. If input/output add it to the appropriate vector.
  //  3. Add it to a name->gate map so that we can wire it up
  //     based on the compoent's name.
  //
  for (int gi=0; gi<numGates; gi++) {
    std::string gateType;
    std::string gateName;

    // Read the gate's type.
    // I.e. one of "and", "or", "not", "input", or "output".
    fs >> gateType;

    // Read the gate's identifying name (for wiring later).
    fs >> gateName;

    std::cerr << "Adding " << gateType << " named '" << gateName << "'.\n";

    // Make a new gate; add it to the data structures.
    //
    std::shared_ptr<Gate> gate;
    if (gateType == "and") {
      gate = std::shared_ptr<And> { new And {gateName} };
    } else if (gateType == "or") {
      gate = std::shared_ptr<Or> {new Or {gateName} };
    } else if (gateType == "not") {
      gate = std::shared_ptr<Not> {new Not {gateName} };
    } else if (gateType == "input") {
      // If an input, also add it to the input list.
      std::shared_ptr<Input> igate = std::shared_ptr<Input> {new Input {gateName} };
      gate = igate;
      inputs.push_back(igate);
    } else if (gateType == "output") {
      // If an output, also add it to the output list.
      std::shared_ptr<Output> ogate = std::shared_ptr<Output> {new Output {gateName} };
      gate = ogate;
      outputs.push_back(ogate);
    }
    // Add it to the circuit's name->gate map so that we can wire things.
    components.insert(std::pair<std::string,std::shared_ptr<Gate>> {gateName,gate});
  }

  // Read each wiring line, linking up circuit components.
  //
  int numWires;
  fs >> numWires;
  
  for (int wi=0; wi<numWires; wi++) {
    std::string wire_tag, from, to;
    fs >> wire_tag; // read "wire"
    fs >> from;
    fs >> to;
    std::cerr << "Wire from '" << from << "' to '" << to << "'.\n";
    std::shared_ptr<Gate> from_g = components[from];
    std::shared_ptr<Gate> to_g = components[to];
    from_g->addOutput(to_g);
    to_g->addInput(from_g);
  }
}

//
// reset
// 
// Prepares each gate so that the circuit can be evaluated.
//
void Circuit::reset() {
  for (auto name_gate : components) {
    name_gate.second->reset();
  }
}

//
// inputSize
//
// Returns the number of inputs into the circuit.
//
int Circuit::inputSize() const {
  return inputs.size();
}

//
// outputSize
//
// Returns the number of outputs from a circuit.
//
int Circuit::outputSize() const {
  return outputs.size();
}

//
// setInputsFrom
//
// Sets each of the Input components according to a boolean
// vector. The order of the vector corresponds to the order of the
// circuit's inputs (the same order that they occur in the file).
//
void Circuit::setInputsFrom(const std::vector<bool>& vals) {
  for (int i = 0; i < inputSize(); i++) {
    inputs[i]->setValue(vals[i]);
  }
}
 
//
// readInputsInto
//
// Reads each of the Outputs components to set the contents of a given
// boolean vector. The order of the vector corresponds to the order of
// the circuit's oututs (the same order that they occur in the file).
//
void Circuit::readOutputsInto(std::vector<bool>& vals) const {
  for (int i = 0; i < outputSize(); i++) {
    vals[i] = outputs[i]->outputValue();
  }
}

// _outputTruthTable
//
// helper functions
//
//
std::string spaces(int n) {
  return std::string(n,' ');
}
//
std::string dashes(int n) {
  return std::string(n,'-');
}
//
void Circuit::_outputHeader() {
  for (auto g : inputs) {
    std::cout << " " << g->name << " |";
  }
  for (auto g : outputs) {
    std::cout << "| " << g->name << " ";
  }
  std::cout << "\n";
  for (auto g : inputs) {
    std::cout << dashes(g->name.length()+2) << "+";
  }
  for (auto g : outputs) {
    std::cout << "+" << dashes(g->name.length()+2);
  }
  std::cout << "\n";
}
//
void Circuit::_outputRow(std::vector<bool> i, std::vector<bool> o) {
  
  for (int c = 0; c < i.size(); c++) {
    std::cout << spaces(inputs[c]->name.length()) << i[c] << " |";
  }
  for (int c = 0; c < o.size(); c++) {
    std::cout << "| " << o[c] << spaces(outputs[c]->name.length());
  }
  std::cout << "\n";
}
//
void advance(std::vector<bool>& bs) {
  auto where = std::find(bs.rbegin(), bs.rend(), false);
  for (auto place = bs.rbegin(); place != where; place++) {
    *place = false;
  }
  if (where != bs.rend()) {
    *where = true;
  }
}
//
bool all_false(std::vector<bool> bs) {
  return std::all_of(bs.begin(),
		     bs.end(),
		     [](bool b)->bool{ return !b; });
}

//
// outputTruthTable
//
// Outputs the full series of rows for the truth table of this
// circuit by repeatedly calling reset/setInputsFrom/evaluate
// and ouputting the results read by readOutputsInto.
//
//
void Circuit::outputTruthTable() {
  std::vector<bool> ins (inputs.size(),false);
  std::vector<bool> outs (outputs.size(),false);
  _outputHeader();
  do {
    reset();
    setInputsFrom(ins);
    evaluate();
    readOutputsInto(outs);

    _outputRow(ins,outs);

    advance(ins);
  } while (!all_false(ins));
}

//
// LogicSim assignment: write the method below!
//

//
// evaluate
//
// Traverses the circuit's components in such a way that the outputs
// get computed based on the values set for each of the inputs.
//
// WRITE THIS CODE!
//
void Circuit::evaluate() {
}

